/*
 * C
 *
 * Copyright 2012-2014 IS2T. All rights reserved.
 * For demonstration purpose only.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 * 10/1/2012 update Controler init
 */

#include "LLDISPLAY_STM32x0GEVAL.h"
#define LLDISPLAY_COPY_IMPL LLDISPLAY_STM32x0GEVAL
#include "LLDISPLAY_COPY_impl.h"
#include "cpu.h"
#if MCU == 2
#include "stm32f2xx.h"
#include "stm322xg_eval_lcd.h"
#include "stm322xg_eval_fsmc_sram.h"
#elif MCU == 4
#include "stm32f4xx.h"
#include "stm324xg_eval_lcd.h"
#include "stm324xg_eval_fsmc_sram.h"
#else
#error Invalid MCU
#endif
#include "microej.h"

typedef struct
{
	__IO uint16_t LCD_REG;
	__IO uint16_t LCD_RAM;
} LCD_TypeDef;

#define LCD_BASE	((uint32_t)(0x60000000 | 0x08000000))
#define LCD_BUFFER	((LCD_TypeDef *) LCD_BASE)
#define WIDTH		LCD_PIXEL_WIDTH
#define HEIGHT		LCD_PIXEL_HEIGHT
#define SRAM_START (0x64000000)
#define SRAM_SIZE (2*1024*1024)
#define MEM_BUFFER	(SRAM_START)
#define BUFFER_SIZE (WIDTH*HEIGHT*16/8)
#define MEM_WORKING_BUFFER_START (MEM_BUFFER + BUFFER_SIZE)


static void _delay_( uint32_t nCount)
{
	uint32_t index = 0;
	for(index = (100000 * nCount); index != 0; index--);
}

/**
 * @brief  Writes to the selected LCD register.
 * @param  LCD_Reg: address of the selected register.
 * @param  LCD_RegValue: value to write to the selected register.
 * @retval None
 */
static inline void writeReg (uint8_t LCD_Reg, uint16_t LCD_RegValue) 
{
	/* Write 16-bit Index, then Write Reg */
	LCD_BUFFER->LCD_REG = LCD_Reg;
	/* Write 16-bit Reg */
	LCD_BUFFER->LCD_RAM = LCD_RegValue;
}

/**
 * @brief  To initialize LCD contoler 9320
 * @param  None
 * @param  None
 * @retval None
 */
void init_9320(void)
{
	/* Start Initial Sequence ------------------------------------------------*/
	writeReg(LCD_REG_229,0x8000); /* Set the internal vcore voltage */
	writeReg(LCD_REG_0,  0x0001); /* Start internal OSC. */
	writeReg(LCD_REG_1,  0x0100); /* set SS and SM bit */
	writeReg(LCD_REG_2,  0x0700); /* set 1 line inversion */
	writeReg(LCD_REG_3,  0x1030); /* set GRAM write direction and BGR=1. */
	writeReg(LCD_REG_4,  0x0000); /* Resize register */
	writeReg(LCD_REG_8,  0x0202); /* set the back porch and front porch */
	writeReg(LCD_REG_9,  0x0000); /* set non-display area refresh cycle ISC[3:0] */
	writeReg(LCD_REG_10, 0x0000); /* FMARK function */
	writeReg(LCD_REG_12, 0x0000); /* RGB interface setting */
	writeReg(LCD_REG_13, 0x0000); /* Frame marker Position */
	writeReg(LCD_REG_15, 0x0000); /* RGB interface polarity */

	/* Power On sequence -----------------------------------------------------*/
	writeReg(LCD_REG_16, 0x0000); /* SAP, BT[3:0], AP, DSTB, SLP, STB */
	writeReg(LCD_REG_17, 0x0000); /* DC1[2:0], DC0[2:0], VC[2:0] */
	writeReg(LCD_REG_18, 0x0000); /* VREG1OUT voltage */
	writeReg(LCD_REG_19, 0x0000); /* VDV[4:0] for VCOM amplitude */
	_delay_(20);                 /* Dis-charge capacitor power voltage (200ms) */
	writeReg(LCD_REG_16, 0x17B0); /* SAP, BT[3:0], AP, DSTB, SLP, STB */
	writeReg(LCD_REG_17, 0x0137); /* DC1[2:0], DC0[2:0], VC[2:0] */
	_delay_(5);                  /* Delay 50 ms */
	writeReg(LCD_REG_18, 0x0139); /* VREG1OUT voltage */
	_delay_(5);                  /* Delay 50 ms */
	writeReg(LCD_REG_19, 0x1d00); /* VDV[4:0] for VCOM amplitude */
	writeReg(LCD_REG_41, 0x0013); /* VCM[4:0] for VCOMH */
	_delay_(5);                  /* Delay 50 ms */
	writeReg(LCD_REG_32, 0x0000); /* GRAM horizontal Address */
	writeReg(LCD_REG_33, 0x0000); /* GRAM Vertical Address */

	/* Adjust the Gamma Curve ------------------------------------------------*/

	writeReg(LCD_REG_48, 0x0007);
	writeReg(LCD_REG_49, 0x0007);
	writeReg(LCD_REG_50, 0x0007);
	writeReg(LCD_REG_53, 0x0007);
	writeReg(LCD_REG_54, 0x0007);
	writeReg(LCD_REG_55, 0x0700);
	writeReg(LCD_REG_56, 0x0700);
	writeReg(LCD_REG_57, 0x0700);
	writeReg(LCD_REG_60, 0x0700);
	writeReg(LCD_REG_61, 0x1F00);

	/* Set GRAM area ---------------------------------------------------------*/
	writeReg(LCD_REG_80, 0x0000); /* Horizontal GRAM Start Address */
	writeReg(LCD_REG_81, 0x00EF); /* Horizontal GRAM End Address */
	writeReg(LCD_REG_82, 0x0000); /* Vertical GRAM Start Address */
	writeReg(LCD_REG_83, 0x013F); /* Vertical GRAM End Address */
	writeReg(LCD_REG_96,  0x2700); /* Gate Scan Line */
	writeReg(LCD_REG_97,  0x0001); /* NDL,VLE, REV */
	writeReg(LCD_REG_106, 0x0000); /* set scrolling line */

	/* Partial Display Control -----------------------------------------------*/
	writeReg(LCD_REG_128, 0x0000);
	writeReg(LCD_REG_129, 0x0000);
	writeReg(LCD_REG_130, 0x0000);
	writeReg(LCD_REG_131, 0x0000);
	writeReg(LCD_REG_132, 0x0000);
	writeReg(LCD_REG_133, 0x0000);

	/* Panel Control ---------------------------------------------------------*/
	writeReg(LCD_REG_144, 0x0010);
	writeReg(LCD_REG_146, 0x0000);
	writeReg(LCD_REG_147, 0x0003);
	writeReg(LCD_REG_149, 0x0110);
	writeReg(LCD_REG_151, 0x0000);
	writeReg(LCD_REG_152, 0x0000);

	/* Set GRAM write direction and BGR = 1 */
	/* I/D=01 (Horizontal : increment, Vertical : decrement) */
	/* AM=1 (address is updated in vertical writing direction) */
	writeReg(LCD_REG_3, 0x1018);

	writeReg(LCD_REG_7, 0x0173); /* 262K color and display ON */
}

/**
 * @brief  To initialize LCD contoler 9325
 * @param  None
 * @param  None
 * @retval None
 */
void init_9325(void)
{
	/* Start Initial Sequence ------------------------------------------------*/
	writeReg(LCD_REG_0, 0x0001); /* Start internal OSC. */
	writeReg(LCD_REG_1, 0x0100); /* Set SS and SM bit */
	writeReg(LCD_REG_2, 0x0700); /* Set 1 line inversion */
	writeReg(LCD_REG_3, 0x1018); /* Set GRAM write direction and BGR=1. */
	writeReg(LCD_REG_4, 0x0000); /* Resize register */
	writeReg(LCD_REG_8, 0x0202); /* Set the back porch and front porch */
	writeReg(LCD_REG_9, 0x0000); /* Set non-display area refresh cycle ISC[3:0] */
	writeReg(LCD_REG_10, 0x0000); /* FMARK function */
	writeReg(LCD_REG_12, 0x0000); /* RGB interface setting */
	writeReg(LCD_REG_13, 0x0000); /* Frame marker Position */
	writeReg(LCD_REG_15, 0x0000); /* RGB interface polarity */

	/* Power On sequence -----------------------------------------------------*/
	writeReg(LCD_REG_16, 0x0000); /* SAP, BT[3:0], AP, DSTB, SLP, STB */
	writeReg(LCD_REG_17, 0x0000); /* DC1[2:0], DC0[2:0], VC[2:0] */
	writeReg(LCD_REG_18, 0x0000); /* VREG1OUT voltage */
	writeReg(LCD_REG_19, 0x0000); /* VDV[4:0] for VCOM amplitude */
	_delay_(20);                      /* Dis-charge capacitor power voltage (200ms) */
	writeReg(LCD_REG_16, 0x17B0); /* SAP, BT[3:0], AP, DSTB, SLP, STB */
	writeReg(LCD_REG_17, 0x0137); /* DC1[2:0], DC0[2:0], VC[2:0] */
	_delay_(5);                       /* Delay 50 ms */
	writeReg(LCD_REG_18, 0x0139); /* VREG1OUT voltage */
	_delay_(5);                       /* Delay 50 ms */
	writeReg(LCD_REG_19, 0x1d00); /* VDV[4:0] for VCOM amplitude */
	writeReg(LCD_REG_41, 0x0013); /* VCM[4:0] for VCOMH */
	_delay_(5);                       /* Delay 50 ms */
	writeReg(LCD_REG_32, 0x0000); /* GRAM horizontal Address */
	writeReg(LCD_REG_33, 0x0000); /* GRAM Vertical Address */

	/* Adjust the Gamma Curve (ILI9325)---------------------------------------*/

	// Style: website 1? ?
	writeReg(0x0030, 0x0000);
	writeReg(0x0031, 0x0707);
	writeReg(0x0032, 0x0307);
	writeReg(0x0035, 0x0200);
	writeReg(0x0036, 0x0008);
	writeReg(0x0037, 0x0004);
	writeReg(0x0038, 0x0000);
	writeReg(0x0039, 0x0707);
	writeReg(0x003C, 0x0002);
	writeReg(0x003D, 0x1D04);

	/* Set GRAM area ---------------------------------------------------------*/
	writeReg(LCD_REG_80, 0x0000); /* Horizontal GRAM Start Address */
	writeReg(LCD_REG_81, 0x00EF); /* Horizontal GRAM End Address */
	writeReg(LCD_REG_82, 0x0000); /* Vertical GRAM Start Address */
	writeReg(LCD_REG_83, 0x013F); /* Vertical GRAM End Address */

	writeReg(LCD_REG_96,  0xA700); /* Gate Scan Line(GS=1, scan direction is G320~G1) */
	writeReg(LCD_REG_97,  0x0001); /* NDL,VLE, REV */
	writeReg(LCD_REG_106, 0x0000); /* set scrolling line */

	/* Partial Display Control -----------------------------------------------*/
	writeReg(LCD_REG_128, 0x0000);
	writeReg(LCD_REG_129, 0x0000);
	writeReg(LCD_REG_130, 0x0000);
	writeReg(LCD_REG_131, 0x0000);
	writeReg(LCD_REG_132, 0x0000);
	writeReg(LCD_REG_133, 0x0000);

	/* Panel Control ---------------------------------------------------------*/
	writeReg(LCD_REG_144, 0x0010);
	writeReg(LCD_REG_146, 0x0000);
	writeReg(LCD_REG_147, 0x0003);
	writeReg(LCD_REG_149, 0x0110);
	writeReg(LCD_REG_151, 0x0000);
	writeReg(LCD_REG_152, 0x0000);

	/* set GRAM write direction and BGR = 1 */
	/* I/D=00 (Horizontal : increment, Vertical : decrement) */
	/* AM=1 (address is updated in vertical writing direction) */
	//writeReg(LCD_REG_3, 0x1018);

	writeReg(LCD_REG_7, 0x0133); /* 262K color and display ON */
}

void init_5408(void)
{
	// Start Initial Sequence --------------------------------------------------
	//			writeReg(LCD_REG_0,   0x0000); // no use?
	writeReg(LCD_REG_1,   0x0100); // set SS bit
	writeReg(LCD_REG_2,   0x0700); // set 1 line inversion
	writeReg(LCD_REG_3,   0x1030); // set GRAM write direction and BGR=1.
	writeReg(LCD_REG_4,   0x0000); // Resize register

	writeReg(LCD_REG_8,  0x0202); // set the back porch and front porch
	writeReg(LCD_REG_9,  0x0000); // set non-display area refresh cycle ISC[3:0]
	writeReg(LCD_REG_10, 0x0000); // FMARK function
	writeReg(LCD_REG_12, 0x0000); // RGB 18-bit System interface setting
	writeReg(LCD_REG_13, 0x0000); // Frame marker Position
	writeReg(LCD_REG_15, 0x0000); // RGB interface polarity, no impact

	// Power On sequence -------------------------------------------------------
	writeReg(LCD_REG_16, 0x0000); // SAP, BT[3:0], AP, DSTB, SLP, STB
	writeReg(LCD_REG_17, 0x0000); // DC1[2:0], DC0[2:0], VC[2:0]
	writeReg(LCD_REG_18, 0x0000); // VREG1OUT voltage
	writeReg(LCD_REG_19, 0x0000); // VDV[4:0] for VCOM amplitude
	_delay_(200);             // Dis-charge capacitor power voltage (200ms)
	writeReg(LCD_REG_16, 0x1690); // SAP, BT[3:0], AP, DSTB, SLP, STB
	writeReg(LCD_REG_17, 0x0221); // ?need update?  DC1[2:0], DC0[2:0], VC[2:0]
	_delay_(50);              // delay 50 ms
	writeReg(LCD_REG_18, 0x001c); // External reference voltage= Vci
	_delay_(50);                 // delay 50 ms
	writeReg(LCD_REG_19, 0x1800); // VDV[4:0] for VCOM amplitude
	writeReg(LCD_REG_41, 0x001c); // VCM[4:0] for VCOMH
	writeReg(LCD_REG_43, 0x000d); // Frame Rate 0x000d = 128Hz(91Hz?), 0x000b = 96Hz
	_delay_(50);                 // delay 50 ms
	writeReg(LCD_REG_32, 0x0000); // GRAM horizontal Address
	writeReg(LCD_REG_33, 0x0000); // GRAM Vertical Address

	// Adjust the Gamma Curve (SPFD5408B)---------------------------------------
	// SPFD5408B_Init.htm, this config seems a bit better than 5408B.INI
	writeReg(LCD_REG_48, 0x000A);
	writeReg(LCD_REG_49, 0x1326);
	writeReg(LCD_REG_50, 0x0A29);
	writeReg(LCD_REG_51, 0x290A);
	writeReg(LCD_REG_52, 0x2613);
	writeReg(LCD_REG_53, 0x0A0A);
	writeReg(LCD_REG_54, 0x1E03);
	writeReg(LCD_REG_55, 0x031E);
	writeReg(LCD_REG_56, 0x0706);
	writeReg(LCD_REG_57, 0x0303);
	writeReg(LCD_REG_58, 0x0E04);
	writeReg(LCD_REG_59, 0x0E01);
	writeReg(LCD_REG_60, 0x010E);
	writeReg(LCD_REG_61, 0x040E);
	writeReg(LCD_REG_62, 0x0303);
	writeReg(LCD_REG_63, 0x0607);

	// Set GRAM area -----------------------------------------------------------
	writeReg(LCD_REG_80, 0x0000); // Horizontal GRAM Start Address
	writeReg(LCD_REG_81, 0x00EF); // Horizontal GRAM End Address
	writeReg(LCD_REG_82, 0x0000); // Vertical GRAM Start Address
	writeReg(LCD_REG_83, 0x013F); // Vertical GRAM End Address

	writeReg(LCD_REG_96,  0xA700); // Gate Scan Line
	writeReg(LCD_REG_97,  0x0001); // NDL,VLE, REV
	writeReg(LCD_REG_106, 0x0000); // set scrolling line

	// Partial Display Control -------------------------------------------------
	writeReg(LCD_REG_128, 0x0000);
	writeReg(LCD_REG_129, 0x0000);
	writeReg(LCD_REG_130, 0x0000);
	writeReg(LCD_REG_131, 0x0000);
	writeReg(LCD_REG_132, 0x0000);
	writeReg(LCD_REG_133, 0x0000);

	// Panel Control -----------------------------------------------------------
	writeReg(LCD_REG_144, 0x0010); // 0x0016?
	writeReg(LCD_REG_146, 0x0000);
	writeReg(LCD_REG_147, 0x0003); //
	writeReg(LCD_REG_149, 0x0110);
	writeReg(LCD_REG_151, 0x0000);
	writeReg(LCD_REG_152, 0x0000);

	// set GRAM write direction and BGR=1
	// I/D=01 (Horizontal : increment, Vertical : decrement)
	// AM=1 (address is updated in vertical writing direction)
	writeReg(LCD_REG_3, 0x1018);

	writeReg(LCD_REG_7, 0x0112); // 262K color and display ON

}

/**
 * @brief  Initializes the LCD.
 * @param  None
 * @retval None
 */
void _STM32xxG_LCD_Init(void)
{

	/* Configure the LCD Control pins --------------------------------------------*/
	LCD_CtrlLinesConfig();

	/* Configure the FSMC Parallel interface -------------------------------------*/
	LCD_FSMCConfig();

	_delay_(5); /* delay 50 ms */

	/* Read the LCD ID */
	switch  ( LCD_ReadReg(0x00))
	{
	/* Check if the LCD is ILI9320 Controller */
	case 0x9320:
	{
		init_9320();
		break;
	}
	case 0x9325: /* Check if the LCD is ILI9325 Controller */
	{
		init_9325();
		break;
	}
	case 0x5408: /* Check if the LCD is SPFD5408B Controller */
	{
		init_5408();
		break;
	}
	}

	LCD_SetFont(&LCD_DEFAULT_FONT);
}


uint8_t LLDISPLAY_COPY_IMPL_initialize(LLDISPLAY_COPY* env)
{
	_STM32xxG_LCD_Init();
	LCD_Clear(White);
	return MICROEJ_TRUE;
}

/*
 * This function has to return the start address of a memory space which will
 * be used to store the working buffer. This buffer is used by the display
 * module to decode dynamically the PNG images.
 *
 * The idea is to use all available remaining SDRAM memory. The SDRAM memory
 * already contains the 2 display buffers (back buffer and LCD buffer).
 */
int32_t LLDISPLAY_COPY_IMPL_getWorkingBufferStartAddress(LLDISPLAY_COPY* env)
{
	return MEM_WORKING_BUFFER_START;
}

int32_t LLDISPLAY_COPY_IMPL_getWorkingBufferEndAddress(LLDISPLAY_COPY* env)
{
	// the working end address is the end address of the available memory (we
	// take all remaining memory to store the working buffer, see
	// LLDISPLAY_COPY_IMPL_getWorkingBufferStartAddress function comment)
	return (int32_t)(SRAM_START + SRAM_SIZE);
}

void LLDISPLAY_COPY_IMPL_backlightOff(LLDISPLAY_COPY* env)
{
	// not available yet
}

void LLDISPLAY_COPY_IMPL_backlightOn(LLDISPLAY_COPY* env)
{
	// not available yet
}

int32_t LLDISPLAY_COPY_IMPL_convertDisplayColorToRGBColor(LLDISPLAY_COPY* env, int32_t color)
{
	//			RED													GREEN													BLUE
	//			XXXXX000.00000000						00000XXX.XXX00000							00000000.000XXXXX
	return ((color & 0xf800) << 8) | ((color & 0x07e0) << 5) | ((color & 0x001f) << 3) | 0xff000000 /* always opaque */;
}

int32_t LLDISPLAY_COPY_IMPL_convertRGBColorToDisplayColor(LLDISPLAY_COPY* env, int32_t color)
{
	uint32_t rgbColor = (uint32_t)color;
	//			RED															GREEN														BLUE
	//			XXXXX000.00000000								00000XXX.XXX00000								00000000.000XXXXX
	return ((rgbColor & 0xf80000) >> 8) | ((rgbColor & 0x00fc00) >> 5) | ((rgbColor & 0xf8) >> 3);
}

void LLDISPLAY_COPY_IMPL_copyBuffer(LLDISPLAY_COPY* env, int32_t xmin, int32_t ymin, int32_t xmax, int32_t ymax)
{
	uint16_t* addr = (uint16_t*)((int32_t)MEM_BUFFER + (ymin * WIDTH * 2));
	int32_t cpt = (ymax - ymin + 1) * WIDTH;	// number of pixels to copy (row aligned)
	LCD_SetCursor(ymin, WIDTH-1);
	LCD_WriteRAM_Prepare();

	// copy pixels
	while(--cpt>=0) {
		LCD_BUFFER->LCD_RAM = *(addr);
		addr++;
	}
}

int32_t LLDISPLAY_COPY_IMPL_getBacklight(LLDISPLAY_COPY* env)
{
	// not available yet
	return 0;
}

int32_t LLDISPLAY_COPY_IMPL_getBackBufferAddress(LLDISPLAY_COPY* env)
{
	// buffer == 0 for sure
	// size == 320*240*16/8 for sure
	//return (int32_t)LCD_buffer;
	return (int32_t)MEM_BUFFER;
}

int32_t LLDISPLAY_COPY_IMPL_getContrast(LLDISPLAY_COPY* env)
{
	// not available yet
	return 0;
}

int32_t LLDISPLAY_COPY_IMPL_getHeight(LLDISPLAY_COPY* env, uint8_t logical)
{
	return HEIGHT;
}

int32_t LLDISPLAY_COPY_IMPL_getWidth(LLDISPLAY_COPY* env, uint8_t logical)
{
	return WIDTH;
}

uint8_t LLDISPLAY_COPY_IMPL_hasBackLight(LLDISPLAY_COPY* env)
{
	return MICROEJ_FALSE;	// not available yet
}

void LLDISPLAY_COPY_IMPL_setBacklight(LLDISPLAY_COPY* env, int32_t backLight)
{
	// not available yet
}

void LLDISPLAY_COPY_IMPL_setContrast(LLDISPLAY_COPY* env, int32_t contrast)
{
	// not available yet
}

void LLDISPLAY_COPY_IMPL_synchronize(LLDISPLAY_COPY* env)
{
	// it is not possible to synchronize the display (missing hardware capacity)
}

int32_t LLDISPLAY_COPY_IMPL_getFontAlphaAdjustment(LLDISPLAY_COPY* env, int32_t bpp)
{
	// '25' has been calibrated manually
    return bpp == 2 ? 25 : 0;
}


uint8_t LLDISPLAY_COPY_IMPL_isColor(LLDISPLAY_COPY* env)
{
	return MICROEJ_TRUE;
}


int32_t LLDISPLAY_COPY_IMPL_getNumberOfColors(LLDISPLAY_COPY* env)
{
	return (1 << 16) - 1;
}

